use std::{error::Error, sync::Arc};

use axum::{
    extract::State,
    response::{Html, IntoResponse},
};
use axum_csrf::CsrfToken;
use axum_macros::debug_handler;
use deadpool_diesel::InteractError;
use diesel::{QueryDsl, RunQueryDsl, SelectableHelper};
use maud::html;
use thiserror::Error;
use tracing::instrument;

use crate::{
    error::AppError,
    models::User,
    partials::{footer, header},
};

#[instrument]
pub async fn index() -> Html<String> {
    Html(
        html! {
            (header())
            main class="prose" {
                h1 { "Manage live lyrics and music displays" }
                p { "Stop struggling to share the same messy set of PowerPoint files or Google Presentations. Make editing and controlling your live lyrics and music displays easy and simple." }
                ul {
                    li { "Live collaboration with your team" }
                    li { "Fully compatible with any device" }
                    li { "Simple workflow" }
                    li { "Dark theme and light theme" }
                    li { "Generous free plan" }
                    li { "Lightweight and fast" }
                }
                section class="flex gap" {
                    a href="/register" class="button bg-primary" { "Try now" }
                    a class="button" href="/login" { "Login" }
                }
            }
            (footer())
        }
        .into_string(),
    )
}

pub async fn register(csrf: CsrfToken) -> impl IntoResponse {
    let token = csrf.authenticity_token().unwrap();
    (
        csrf,
        Html(
            html! {
                (header())
                main class="prose" {
                    h1 { "Register an account" }
                    form method="post" {
                        input type="hidden" name="authenticity_token" value=(token) {}
                        label {
                            "Username:"
                            input name="username" {}
                        }
                        label {
                            "Password:"
                            input type="password" name="password" {}
                        }
                        button type="submit" { "Create Account" }
                    }
                }
                (footer())
            }
            .into_string(),
        ),
    )
        .into_response()
}

pub async fn login(csrf: CsrfToken) -> impl IntoResponse {
    let token = csrf.authenticity_token().unwrap();
    (
        csrf,
        Html(
            html! {
                (header())
                main class="prose" {
                    h1 { "Login" }
                    form method="post" {
                            input type="hidden" name="authenticity_token" value=(token) {}
                        label {
                            input {}
                        }
                        label {
                            input {}
                        }
                    }
                }
                (footer())
            }
            .into_string(),
        ),
    )
        .into_response()
}

#[derive(Error, Debug)]
enum AllUsersError {
    #[error("other application error")]
    App(#[from] AppError),
    #[error("failed to retrieve users")]
    DB(#[from] InteractError),
}

#[debug_handler]
pub async fn all_users(
    State(state): State<crate::state::State>,
) -> Result<Html<String>, AllUsersError> {
    use crate::schema::users::dsl::*;

    let conn = state.database.pool.get().await?;
    let cc = |c| -> Vec<User> {
        users
            .select(User::as_select())
            .load(c)
            .expect("error loading users")
    };

    let all_users: Vec<User> = match conn.interact(cc).await {
        Ok(u) => u,
        Err(e) => return Err(("failed to retrieve users")),
    };

    // @if let Some(name) = u.name {
    //     name
    // } @else {
    //     "N/A"
    // }
    Ok(Html(
        html! {
            (header())
            main class="prose" {
                h1 { "Users" }
                ul {
                    @for u in all_users {
                        li {
                            (u)
                            "("
                            ")"
                        }
                    }
                }
            }
            (footer())
        }
        .into_string(),
    ))
}

#[instrument]
pub async fn greet_world() -> Html<String> {
    Html(
        html! {
            h1 { (greet_world_text().await) }
        }
        .into_string(),
    )
}

#[instrument]
pub async fn greet_world_text() -> &'static str {
    "Hello, World!"
}
